(*
  Copyright 2010 Gerd Stolpmann

  This file is part of Plasma, a distributed filesystem and a
  map/reduce computation framework. Unless you have a written license
  agreement with the copyright holder (Gerd Stolpmann), the following
  terms apply:

  Plasma is free software: you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation, either version 3 of the License, or
  (at your option) any later version.

  Plasma is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with Foobar.  If not, see <http://www.gnu.org/licenses/>.

*)
(* $Id: mapred_taskfiles.ml 283 2010-10-28 00:19:50Z gerd $ *)

open Printf

class type taskfile_manager =
object
  method local_directory : string
  method log_directory : string
  method copy_to_local_directory : string list -> unit
  method delete_local_directory : unit -> unit
  method task_save_log_directory : Plasma_client.plasma_cluster -> string list -> unit
  method task_files_log_directory : unit -> string list
  method task_delete_log_directory : unit -> unit
end


let scp_taskfile_manager mc jc =
  let local_dir =
    mc#mr_task_tmpdir ^ "/" ^ jc#job_id ^ "_" ^ "local" in
  let log_dir =
    mc#mr_task_tmpdir ^ "/" ^ jc#job_id ^ "_" ^ "log" in
  ( object
      method local_directory = local_dir
      method log_directory = log_dir

      method copy_to_local_directory files =
	List.iter
	  (fun f ->
	     if not (Sys.file_exists f) then
	       failwith ("File does not exist: " ^ f)
	  )
	  files;
	List.iter
	  (fun node ->
	     (* Ensure the local directory exists *)
	     let cmd = 
	       sprintf "ssh -q %s mkdir -p %s"
		 (Filename.quote node)
		 (Filename.quote local_dir) in
	     let code = Sys.command cmd in
	     if code <> 0 then
	       failwith ("Exec failed: " ^ cmd);
	     (* Copy the files to the node: *)
	     let files' =
	       List.map
		 (fun f -> if f <> "" && f.[0] <> '/' then "./" ^ f else f)
		 files in
	     let cmd =
	       sprintf "scp -q -p %s %s:%s/"
		 (String.concat " "
		    (List.map Filename.quote files'))
		 (Filename.quote node)
		 (Filename.quote local_dir) in
	     let code = Sys.command cmd in
	     if code <> 0 then
	       failwith ("Exec failed: " ^ cmd);
	  )
	  mc#mr_task_nodes

      method delete_local_directory () =
	assert (local_dir <> "/");
	List.iter
	  (fun node ->
	     let cmd = 
	       sprintf "ssh -q %s rm -rf %s"
		 (Filename.quote node)
		 (Filename.quote local_dir) in
	     let code = Sys.command cmd in
	     if code <> 0 then
	       failwith ("Exec failed: " ^ cmd);
	  )
	  mc#mr_task_nodes

      method task_save_log_directory c files =
	let hostname = Unix.gethostname() in
	let abs_files =
	  List.map
	    (fun f ->
	       if f <> "" && f.[0] = '/' then (f,f) else
		 (f, log_dir ^ "/" ^ f)
	    )
	    files in
	List.iter
	  (fun (_,f) ->
	     if not (Sys.file_exists f) then
	       failwith ("File does not exist: " ^ f)
	  )
	  abs_files;
	(* First ensure that the PlasmaFS log directory exists *)
	let trans = Plasma_client.start c in
	( try
	    let _dir_inode =
	      Plasma_client.retry c "mkdir"
		(Plasma_client.mkdir
		   trans jc#log_dir) (Plasma_client.dir_ii c 0o777) in
	    ()
	  with
	    | Plasma_client.Plasma_error `eexist -> ()
	);
	Plasma_client.commit trans;
	(* Copy file by file *)
	List.iter
	  (fun (f,abs) ->
	     let p_name = jc#log_dir ^ "/" ^ hostname ^ "_" ^ f in
	     let trans = Plasma_client.start c in
	     let inode = 
	       Plasma_client.retry c "create_file"
		 (Plasma_client.create_file
		    trans p_name) (Plasma_client.regular_ii c 0o666) in
	     Plasma_client.commit trans;
	     let st = Unix.LargeFile.stat abs in
	     if st.Unix.LargeFile.st_kind = Unix.S_REG then (
	       let fd = Unix.openfile abs [Unix.O_RDONLY] 0 in
	       try
		 let eof = st.Unix.LargeFile.st_size in
		 Plasma_client.copy_in c inode 0L fd eof `Star;
		 Unix.close fd
	       with
		 | error -> Unix.close fd; raise error
	     )
	  )
	  abs_files;
	(* Finally delete files *)
	List.iter
	  (fun (_,f) ->
	     Sys.remove f
	  )
	  abs_files;
	
      method task_files_log_directory () =
	let files =
	  try
	    Array.to_list (Sys.readdir log_dir)
	  with Sys_error _ -> [] in
	List.filter
	  (fun name ->
	     let st = Unix.stat (log_dir ^ "/" ^ name) in
	     st.Unix.st_kind = Unix.S_REG
	  )
	  files

      method task_delete_log_directory () =
	let cmd =
	  sprintf "rm -rf %s" (Filename.quote log_dir) in
	let code = Sys.command cmd in
	if code <> 0 then
	  failwith ("Exec failed: " ^ cmd)	

    end
  )
